import typing
from typing import List

import pyarrow
import sqlglot
from sqlglot import Expression
from sqlglot.expressions import TableSample

from universql.plugin import UniversqlPlugin, register, UQuery, Locations
from universql.warehouse.duckdb import DuckDBExecutor
from universql.warehouse.snowflake import SnowflakeExecutor


# when FILES is not defined
# COPY INTO stg_device_metadata FROM 's3:/test/initial_objects/*' (TYPE = CSV SKIP_HEADER = 1)
# when FILES is specified:
# COPY INTO stg_device_metadata FROM 's3:/test/initial_objects/device_metadata.csv' (TYPE = CSV SKIP_HEADER = 1)
# COPY INTO stg_device_metadata FROM 's3:/test/initial_objects/file2.csv' (TYPE = CSV SKIP_HEADER = 1)

class StageTransformer(UQuery):
    def transform_ast(self, expression: Expression, target_executor: DuckDBExecutor) -> Expression:
        if isinstance(expression, sqlglot.exp.Var) and expression.name.startswith('@'):
            expression.args['name'] = 'myname'
            return expression

        # referenced from copy
        if isinstance(expression, sqlglot.exp.Table) and expression.alias_or_name.startswith('@'):
            self._get_stage(expression)

        return expression

    def _get_stage(self, table: sqlglot.exp.Table):
        # self.source_executor.execute_raw("DESCRIBE STAGE {}", self.source_executor)
        return

    def post_execute(self, locations: typing.Optional[Locations], target_executor: DuckDBExecutor):
        pass

    def end(self, table : pyarrow.Table):
        pass

@register()
class SnowflakeStageUniversqlPlugin(UniversqlPlugin):
    def __init__(self, session: "universql.protocol.session.UniverSQLSession"):
        super().__init__(session)

    def start_query(self, ast: typing.Optional[List[sqlglot.exp.Expression]], raw_query: str) -> UQuery:
        return StageTransformer(ast, raw_query)


# @register()
class TableSampleUniversqlPlugin(UniversqlPlugin):
    def __init__(self, source_executor: SnowflakeExecutor):
        super().__init__(source_executor)

    def transform_sql(self, expression: Expression, target_executor: DuckDBExecutor,
                      sample="reservoir(1000)") -> Expression:
        if not isinstance(target_executor, DuckDBExecutor):
            raise NotImplementedError
        if isinstance(expression, sqlglot.exp.Select) and 'sample' not in expression.args:
            sample_value = sqlglot.exp.maybe_parse(sample)
            expression.args['sample'] = sqlglot.exp.TableSample(
                method=sqlglot.exp.Var(this=sample_value.this),
                size=sample_value.expressions[0])
        return expression



from contextlib import suppress
from copy import deepcopy
from traceback import print_exc

from fastapi import FastAPI

from universql.plugin import register


# @register()
def run_marimo(app: FastAPI):
    with suppress(ImportError):
        import marimo
        app.mount("/", marimo.create_asgi_app(include_code=False)
                  .with_dynamic_directory(path="/", directory="/Users/bkabak/.universql/marimos")
                  .build())

def configure(components, config):
    # TODO: generalize to arbitrary nested dictionaries, not just one level
    _components = deepcopy(components)
    for k1, v1 in config.items():
        for k2, v2 in v1.items():
            _components[k1][k2] = v2
    return _components


# @register()
async def run_jupyter(app: FastAPI):
    try:
        from asphalt.core import Context
        from jupyverse_api.main import JupyverseComponent
        from jupyverse_api.app import App

        components = configure({"app": {"type": "app"}}, {"app": {"mount_path": '/notebook'}})
        async with Context() as ctx:
            component = JupyverseComponent(components=components, app=app, debug=True)
            component.start(ctx)
            jupyter_app = await ctx.request_resource(App)
            app.mount("/notebook", jupyter_app)
    except Exception as e:
        print_exc(10)
        print(e)




import functools
import inspect

import pyarrow
from fastapi import FastAPI
from sqlglot import Expression

import typing
from abc import ABC, abstractmethod
from typing import List

import pyiceberg.table
import sqlglot

Locations = typing.Dict[sqlglot.exp.Table, sqlglot.exp.Expression | None]
Tables = typing.Dict[sqlglot.exp.Table, pyiceberg.table.Table | None]


class ICatalog(ABC):
    def __init__(self, session: "universql.protocol.session.UniverSQLSession", compute: dict):
        self.context = session.context
        self.session_id = session.session_id
        self.credentials = session.credentials
        self.compute = compute
        self.iceberg_catalog = session.iceberg_catalog

    @abstractmethod
    def get_table_paths(self, tables: List[sqlglot.exp.Table]) -> Tables:
        pass

    @abstractmethod
    def register_locations(self, tables: Locations):
        pass

    @abstractmethod
    def executor(self) -> "Executor":
        pass


T = typing.TypeVar('T', bound=ICatalog)


def _track_call(method):
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        # Mark that the method was called on this instance.
        self._warm = True
        return method(self, *args, **kwargs)

    return wrapper


class Executor(typing.Protocol[T]):

    def __init__(self, catalog: T):
        self.catalog = catalog

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.execute = _track_call(cls.execute)
        cls.execute_raw = _track_call(cls.execute_raw)

    def is_warm(self):
        return getattr(self, '_warm', False)

    @abstractmethod
    def execute(self, ast: sqlglot.exp.Expression, catalog_executor: "Executor", locations: Tables) -> \
            typing.Optional[Locations]:
        pass

    def test(self):
        self.execute_raw("select 1", None)

    @abstractmethod
    def execute_raw(self, raw_query: str, catalog_executor: typing.Optional["Executor"]) -> None:
        pass

    @abstractmethod
    def get_as_table(self) -> pyarrow.Table:
        pass

    @abstractmethod
    def get_query_log(self, total_duration) -> str:
        pass

    @abstractmethod
    def close(self):
        pass


class UQuery:
    def __init__(self, ast: typing.Optional[List[sqlglot.exp.Expression]], raw_query: str):
        self.ast = ast
        self.raw_query = raw_query

    def transform_ast(self, ast: sqlglot.exp.Expression, target_executor: Executor) -> Expression:
        return ast

    def post_execute(self, locations: typing.Optional[Locations], target_executor: Executor):
        pass

    def end(self, table : pyarrow.Table):
        pass


class UniversqlPlugin(ABC):
    def __init__(self,
                 session: "universql.protocol.session.UniverSQLSession"
                 ):
        self.session = session

    def start_query(self, ast: typing.Optional[List[sqlglot.exp.Expression]], raw_query : str) -> UQuery:
        return UQuery(ast, raw_query)


# {"duckdb": DuckdbCatalog ..}
COMPUTES = {}
# [method]
PLUGINS = []
# apps to be installed
APPS = []


def register(name: typing.Optional[str] = None):
    """
    Decorator to register a Compute subclass with an optional name.
    :param name: Unique of the catalog
    :param executor: The optional executor class for the catalog
    """

    def decorator(cls):
        if inspect.isclass(cls):
            if issubclass(cls, ICatalog) and cls is not ICatalog:
                if name is None:
                    raise SystemError("name is required for catalogs")
                COMPUTES[name] = cls
            elif issubclass(cls, UniversqlPlugin) and cls is not UniversqlPlugin:
                PLUGINS.append(cls)
        elif inspect.isfunction(cls):
            signature = inspect.signature(cls)
            if len(signature.parameters) == 1 and signature.parameters.values().__iter__().__next__().annotation is FastAPI:
                APPS.append(cls)
        else:
            raise SystemError(f"Unknown type {cls}")
        return cls

    return decorator



import typing
from typing import List

import sqlglot
from duckdb.experimental.spark.errors import UnsupportedOperationException
from snowflake.connector.options import pyarrow

from universql.protocol.session import UniverSQLSession
from universql.plugin import ICatalog, Executor, Locations, Tables, register


@register(name="redshift")
class RedshiftCatalog(ICatalog):
    def __init__(self, session: UniverSQLSession, compute: dict):
        super().__init__(session, compute)

    def get_table_paths(self, tables: List[sqlglot.exp.Table]) -> Tables:
        raise UnsupportedOperationException("BigQuery does not support registering tables")

    def register_locations(self, tables: Locations):
        pass

    def executor(self) -> Executor:
        return RedshiftExecutor(self)

class RedshiftExecutor(Executor):
    def __init__(self, catalog: RedshiftCatalog):
        super().__init__(catalog)

    def execute(self, ast: sqlglot.exp.Expression, catalog_executor: Executor, locations: Tables) -> typing.Optional[Locations]:
        return None

    def execute_raw(self, raw_query: str, catalog_executor: Executor) -> None:
        pass

    def get_as_table(self) -> pyarrow.Table:
        pass

    def get_query_log(self, total_duration) -> str:
        pass

    def close(self):
        pass
